#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-
import copy
import datetime
import json
import logging
import requests
import time
import numpy as np
from pytz import timezone
from mqtt_interface import MqttInterface
from object_info import PerceptibleObject

FORMAT = '%(asctime)s-%(levelname)s: %(message)s'
logging.basicConfig(level=logging.INFO, format=FORMAT)
logger = logging.getLogger(__name__)


class LayoutMonitoring():

    def __init__(self):
        service_host = "xxx.xxx.xxx.xxx" # set
        service_port = 00000 # set
        self._request_api_url = "http://RDBS/rdr/api/get/layoutInfo" # set
        self._threshold_path = "./config/layout_threshold_conf.json"
        self._sampling_dict = {}
        self.mptt_req_res_ins = MqttInterface(
            service_host, service_port,
            pub_topic="obj_recognition_01/request_layout",
            sub_topic="obj_recognition_01/response_layout")
        self.mptt_pub_ins = MqttInterface(
            service_host, service_port,
            pub_topic="obj_recognition_01/changed_layout")
        self._SAMPLING_MAX = 30
        self._RETRY_COUNT_MAX = 3
        self._HTTP_RESPONSE_WAIT_SEC = 1
        self._HTTP_RESPONSE_OK = 200

    def sampling_run(self, object_list_a, target_class):
        """
        Perform sampling of the target quasi-static object class

        Parameters
        ----------
        object_list_a: Object-recognized object list
        target_class: Sampling target class

        Returns
        ----------
        List of sampled target class objects
        """
        # Store only the data of the target class in the array
        table_a = []
        result = False
        for dtc_obj in object_list_a:
            if dtc_obj.recognition_name == target_class:
                table_a.append(dtc_obj)
        if self._sampling_dict.get(target_class) is None:
            self._sampling_dict[target_class] = []

        # Add to the array until the number of rows is 30 or more
        self._sampling_dict[target_class].append(table_a)
        if len(self._sampling_dict[target_class]) < self._SAMPLING_MAX:
            return result, []

        logger.info("sampling running")
        # Take out the first data and sort it
        sort_table_a = self.sort_coordinate(self._sampling_dict[target_class].pop(0))

        # Set the sorted data as comparison data
        comparison_a = copy.copy(sort_table_a)
        sum_comp_dict = {
            "sum_coord": [np.array(target.world_coordinate) for target in copy.copy(sort_table_a)],
            "sum_count": [1 for x in range(len(sort_table_a))]
        }
        sampling_sort_a = [sort_table_a]
        for sampling_table_a in self._sampling_dict[target_class]:
            # Creating a row table similar to the comparison data
            sampling_row_a = self.match_sampling_rows(sampling_table_a, comparison_a)
            sampling_sort_a.append(sampling_row_a)
            # Update table for comparison
            comparison_a, sum_comp_dict = self.comp_update(sampling_row_a, comparison_a, sum_comp_dict)
        # Initialize the sampling data storage array
        self._sampling_dict[target_class] = []
        # Flatten an irregular multidimensional table
        flat_sampling_a = self.flattening_table(sampling_sort_a)
        sampling_average_a = self.create_sampling_average(flat_sampling_a)
        result = True
        return result, sampling_average_a

    def comp_update(self, sampling_a, comparison_a, sum_comp_dict):
        """
        Update the table data used for comparison

        Parameters
        ----------
        sampling_a: sampling data
        comparison_a: comparison data
        sum_comp_dict: Dictionary data for calculating the average

        Returns
        ----------
        comparison_a: update comparison data
        sum_comp_dict: Dictionary data for calculating the updated average
        """
        for index in range(len(sampling_a)):
            # Check the number of data for comparison
            if sampling_a[index] is not None:
                # Update the comparison data if it is less than the number of comparison data
                if index < len(comparison_a):
                    sum_comp_dict["sum_coord"][index] += np.array(sampling_a[index].world_coordinate)
                    sum_comp_dict["sum_count"][index] += 1
                    # Update if there is the same position data
                    comparison_a[index] = copy.copy(sampling_a[index])
                    denominator = float(sum_comp_dict["sum_count"][index])
                    world_coordinate = sum_comp_dict["sum_coord"][index] / np.array([denominator] * 3)
                    comparison_a[index].world_coordinate = world_coordinate.tolist()
                else:
                    # Add data to the table when the number of data for comparison is small
                    sum_comp_dict["sum_coord"].append(np.array(sampling_a[index].world_coordinate))
                    sum_comp_dict["sum_count"].append(1)
                    # Add newly added position data to comparison table data
                    comparison_a.append(sampling_a[index])
            else:
                # Do not update if there is no same position data
                pass
        return comparison_a, sum_comp_dict

    def match_sampling_rows(self, sampling_table_a, comparison_a):
        """
        Create sampling row data in a sequence that matches the comparison data

        Parameters
        ----------
        sampling_a: sampling data
        comparison_a: comparison data

        Returns
        ----------
        comparison_a: update comparison data
        sum_comp_dict: Dictionary data for calculating the updated average
        """
        # Define List to store sorted data
        match_table_a = [None for x in range(len(comparison_a))]
        regster_comparison_index_a = []
        for sampling_data in sampling_table_a:
            similar_data = None
            similar_index = None
            for comp_data in comparison_a:
                # Compare the data to determine if they are the same position data
                if self.layout_comparison(comp_data.recognition_name, comp_data, sampling_data):
                    regster_check_index = comparison_a.index(comp_data)
                    if regster_check_index in regster_comparison_index_a:
                        pass
                    elif similar_data is None:
                        # Data is stored when it is judged to be the same position
                        similar_data = copy.copy(sampling_data)
                        similar_index = comparison_a.index(comp_data)
                    else:
                        # Select closer data if data is already stored
                        if self.comparison_differences(comp_data.world_coordinate,
                                                       similar_data.world_coordinate,
                                                       sampling_data.world_coordinate):
                            similar_data = copy.copy(sampling_data)
                            similar_index = comparison_a.index(comp_data)
            if similar_data is not None:
                # If there is the same position data, store it in the corresponding position
                match_table_a[similar_index] = copy.copy(similar_data)
                regster_comparison_index_a.append(similar_index)
            else:
                # If there is no same position data, add it as new data
                match_table_a.append(sampling_data)
        return match_table_a

    def sort_coordinate(self, sort_table_a):
        """
        Sorts the estimated position data table.

        Parameters
        ----------
        sort_table_a : Table array to sort

        Returns
        ----------
        Sorted table array.
        """
        for i in range(len(sort_table_a)):
            for j in range(len(sort_table_a) - 1 - i):
                if sort_table_a[j].world_coordinate[0] > sort_table_a[j + 1].world_coordinate[0]:
                    temp = copy.copy(sort_table_a[j])
                    sort_table_a[j] = copy.copy(sort_table_a[j + 1])
                    sort_table_a[j + 1] = copy.copy(temp)
        return sort_table_a

    def flattening_table(self, irregular_table):
        """
        Make the number of elements in a multidimensional array uniform

        Parameters
        ----------
        irregular_table : Multidimensional array containing discrete elements

        Returns
        ----------
        Flattened multidimensional array.
        """
        # Get the maximum number of elements
        elem_max = max(map(len, irregular_table))
        for row_a in irregular_table:
            # Determine if the number of elements is small
            if len(row_a) < elem_max:
                duff_range = elem_max - len(row_a)
                empty_list = [None for i in range(0, duff_range)]
                # Add None for the number of missing elements
                row_a.extend(empty_list)
        return irregular_table

    def create_sampling_average(self, sampling_tbl_a):
        """
        Calculate the average value of the sampling data for each column

        Parameters
        ----------
        sampling_tbl_a : sampling data table

        Returns
        ----------
        object list with average value.
        """
        avg_list_a = []
        # Get the number of columns
        col_max = max(map(len, sampling_tbl_a))

        criteria_count = int((self._SAMPLING_MAX * 2) / 3)
        for col in range(0, col_max):
            object_avg = PerceptibleObject()
            sampling_count = 0
            for row_a in sampling_tbl_a:
                if row_a[col] is None:
                    continue
                if object_avg.recognition_name is None:
                    sampling_count = 1
                    object_avg.recognition_name = row_a[col].recognition_name
                    object_avg.detect_accuracy = row_a[col].detect_accuracy
                    object_avg.image_coordinate = row_a[col].image_coordinate
                    object_avg.world_coordinate = row_a[col].world_coordinate
                else:
                    sampling_count += 1
                    object_avg.detect_accuracy += row_a[col].detect_accuracy
                    image_coordinate = (np.array(object_avg.image_coordinate) + np.array(row_a[col].image_coordinate))
                    object_avg.image_coordinate = image_coordinate.tolist()
                    world_coordinate = (np.array(object_avg.world_coordinate) + np.array(row_a[col].world_coordinate))
                    object_avg.world_coordinate = world_coordinate.tolist()
            # Determine if the number of detections is 2/3 or more
            if criteria_count <= sampling_count:
                # Calculate the average value of sampling data
                object_avg.detect_accuracy /= sampling_count
                image_coordinate = np.array(object_avg.image_coordinate) / np.array([sampling_count, sampling_count])
                object_avg.image_coordinate = image_coordinate.tolist()
                world_coordinate = np.array(object_avg.world_coordinate) / np.array([sampling_count, sampling_count, sampling_count])
                object_avg.world_coordinate = world_coordinate.tolist()
                # Add sampling data to average list
                avg_list_a.append(object_avg)
        return avg_list_a

    def layout_comparison(self, class_name, source_data, targets_data):
        """
        Perform layout comparisons.
        Returns True if within the threshold range,
        Set False outside the threshold range

        Parameters
        ----------
        class_name  : Class to be compared
        source_data : Object data to compare
        targets_data: Target object data to compare

        Returns
        ----------
        Returns the comparison result.
        """
        thr_x = 0
        thr_y = 0
        result = False
        try:
            # Set the path of the configuration file
            config_path = self._threshold_path
            # Open config file
            with open(config_path) as read_json:
                load_json = json.load(read_json)
                for obj in load_json:
                    # Get the threshold of the target class
                    if class_name == obj["class"]:
                        thr_x = obj["threshold_x"]
                        thr_y = obj["threshold_y"]
        except json.JSONDecodeError as e:
            logger.error(e)
            return result
        except OSError as e:
            logger.error(e)
            return result
        comp_x = source_data.world_coordinate[0]
        comp_y = source_data.world_coordinate[1]
        trgt_x = targets_data.world_coordinate[0]
        trgt_y = targets_data.world_coordinate[1]
        # Threshold judgment implementation
        if (comp_x - thr_x) <= trgt_x <= (comp_x + thr_x):
            if (comp_y - thr_y) <= trgt_y <= (comp_y + thr_y):
                result = True
        return result

    def layout_change_detection(self, object_pos_list_a):
        """
        Process to detect layout change

        Parameters
        ----------
        object_pos_list_a: Estimated position list of objects detected by object recognition

        Returns
        ----------
        Returns the comparison result.
        """
        result = False
        # Get the last updated layout data from DB
        request_res, rdr_layout_a = self.request_get_layout()

        if not request_res:
            logger.error("Unable to get layout data")
            return result
        # Sampling for each quasi-static object
        sampling_t_res, sampling_layout_t_a = self.sampling_run(object_pos_list_a, "Table")
        sampling_c_res, sampling_layout_c_a = self.sampling_run(object_pos_list_a, "chair")

        if not sampling_t_res:
            return result
        # Combine sampling data
        sampling_layout_a = sampling_layout_t_a + sampling_layout_c_a
        # Compare the number of layouts and decide on updates
        if len(rdr_layout_a) != len(sampling_layout_a):
            # Updated layout data
            send_data = self.create_send_message(sampling_layout_a)
            if self.mptt_pub_ins.publish_send(send_data):
                result = True
                return result
            else:
                logger.error("Unable to update layout data")
                return result
        sampling_tmp_a = copy.copy(sampling_layout_a)
        for rdr_data in rdr_layout_a:
            similar_data = None
            similar_index = None
            for sampling_data in sampling_tmp_a:
                # Is it the same class name?
                if sampling_data.recognition_name == rdr_data.recognition_name:
                    # Compare positions and judge if they are the same layout
                    if self.layout_comparison(sampling_data.recognition_name, sampling_data, rdr_data):
                        if similar_data is None:
                            # Stores data judged to be the same layout
                            similar_data = copy.copy(sampling_data)
                            similar_index = sampling_tmp_a.index(sampling_data)
                        else:
                            # Determine if the data is more similar
                            if self.comparison_differences(rdr_data.world_coordinate,
                                                           similar_data.world_coordinate,
                                                           sampling_data.world_coordinate):
                                similar_data = copy.copy(sampling_data)
                                similar_index = sampling_tmp_a.index(sampling_data)
            # Is it the same position data?
            if similar_data is not None:
                # Delete the corresponding element.
                sampling_tmp_a.pop(similar_index)
            else:
                break
        # Check the number of newly detected layouts
        if 0 < len(sampling_tmp_a):
            # Updated layout data
            logger.info("Updated layout data")
            send_data = self.create_send_message(sampling_layout_a)
            if self.mptt_pub_ins.publish_send(send_data):
                result = True
            else:
                logger.error("Unable to update layout data")
        return result

    def comparison_differences(self, base_coord, exists_coord, subject_coord):
        """
        Determine if the target data is closer than the already registered data

        Parameters
        ----------
        base_coord: Reference estimated position
        exists_coord: Existing estimated position
        subject_coord: Subject estimated position

        Returns
        ----------
        Returns True if data closer to the criteria is found
        """
        result = False
        diff1 = (base_coord[0] - exists_coord[0]) ** 2
        diff1 += (base_coord[1] - exists_coord[1]) ** 2
        diff2 = (base_coord[0] - subject_coord[0]) ** 2
        diff2 += (base_coord[1] - subject_coord[1]) ** 2
        if diff1 > diff2:
            result = True
        return result

    def create_send_message(self, layout_data_a):
        """
        Convert layout data to dictionary type of
        transmission data format

        Parameters
        ----------
        layout_data_a: Layout data list

        Returns
        ----------
        Message data for sending
        """
        object_data_a = []
        for object_data in layout_data_a:
            object_dict = {
                "object_name": object_data.recognition_name,
                "avg_detection_rate": float(object_data.detect_accuracy),
                "avg_image_x": float(object_data.image_coordinate[0]),
                "avg_image_y": float(object_data.image_coordinate[1]),
                "avg_object_x": float(object_data.world_coordinate[0]),
                "avg_object_y": float(object_data.world_coordinate[1]),
                "avg_object_z": float(object_data.world_coordinate[2])
            }
            object_data_a.append(object_dict)
        message = {
            "robot_id": "object_layout_01",
            "layout_info": {
                "layout_date": (datetime.datetime.now(timezone('UTC')).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]) + "+09:00"
            },
            "object_list": object_data_a,
        }
        return message

    def request_get_layout(self):
        """
        Get layout data via MQTT and convert to structure list

        Returns
        ----------
        result: Execution result
        layout_data_a: Layout data list
        """
        result = False
        layout_data_a = []
        try:
            for retry_count in range(self._RETRY_COUNT_MAX):
                post_res = requests.post(self._request_api_url, timeout=(5.0))
                if post_res.status_code == self._HTTP_RESPONSE_OK:
                    res_dict = json.loads(post_res.text)
                    # Send layout data acquisition request to MQTT
                    layout_data_a = []
                    # Store data in object structure list
                    for dict_data in res_dict["object_list"]:
                        sto_ins = PerceptibleObject()
                        sto_ins.recognition_name = dict_data["object_name"]
                        sto_ins.detect_accuracy = dict_data["avg_detection_rate"]
                        sto_ins.image_coordinate = [dict_data["avg_image_x"],
                                                    dict_data["avg_image_y"]]
                        sto_ins.world_coordinate = [dict_data["avg_object_x"],
                                                    dict_data["avg_object_y"],
                                                    dict_data["avg_object_z"]]
                        layout_data_a.append(sto_ins)
                    result = True
                    break
                else:
                    time.sleep(self._HTTP_RESPONSE_WAIT_SEC)
        except json.JSONDecodeError:
            pass
        except OSError:
            pass
        return result, layout_data_a
