#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-
import json
import logging
import paho.mqtt.client as mqtt
import ssl
import threading
import time

FORMAT = '%(asctime)s-%(levelname)s: %(message)s'
logging.basicConfig(level=logging.INFO, format=FORMAT)
logger = logging.getLogger(__name__)


class MqttInterface(mqtt.Client):

    def __init__(self, set_host, set_port, pub_topic='', sub_topic=''):
        super(MqttInterface, self).__init__()
        self._pub_topic = pub_topic
        self._sub_topic = sub_topic
        self._host = set_host
        self._port = set_port
        self._cacert = './cert/inov_ca.crt'
        self._clientCert = './cert/eac.crt'
        self._clientKey = './cert/eac.key'
        self._sub_connect_flag = False
        self._sub_receive_flag = False
        self._sub_receive_message = None
        self._lock = threading.Lock()
        self._RETRY_CONNECT_MAX = 3
        self._RETRY_PUBLISH_MAX = 3
        self._RETRY_PUBSUB_MAX = 5
        self._CONNECT_WAIT_MAX = 3
        self._CONNECT_WAIT = 1
        self._PUBLISH_WAIT = 5
        self._PUBSUB_WAIT = 1

    def connection(self):
        """
        Connect to MQTT Server

        Returns
        ----------
        result: Connection result
        """
        # Connection initial setting
        result = False
        self._sub_connect_flag = False
        retry_count = 0
        try:
            # Retry processing
            while not self._sub_connect_flag and retry_count < self._RETRY_CONNECT_MAX:
                self.tls_set(self._cacert, certfile=self._clientCert, keyfile=self._clientKey, cert_reqs=mqtt.ssl.CERT_REQUIRED, tls_version=ssl.PROTOCOL_TLSv1_2)
                # Connection
                self.connect(self._host, self._port, 60)
                # Perform subscription processing when there is a topic
                if self._sub_topic:
                    self.subscribe(self._sub_topic)
                # loop start processing
                self.loop_start()
                # Connection result confirmation process
                for wait_count in range(self._CONNECT_WAIT_MAX):
                    time.sleep(self._CONNECT_WAIT)
                    if self._sub_connect_flag:
                        result = True
                        break
                else:
                    # Retry count addition
                    retry_count += 1
        except Exception as e:
            logger.error(e)
        if not result:
            logger.error("Unable to connection for MQTT server")
        return result

    def publish_send(self, message):
        """
        Send a public message

        Parameters
        ----------
        message: Publish message

        Returns
        ----------
        result: Execution result
        """
        result = False
        # Convert message to Json format
        json_message = json.dumps(message)
        for retry_count in range(self._RETRY_PUBLISH_MAX):
            # Publish message
            res = self.publish(self._pub_topic, json_message)
            # Result judgment
            if res[0] == mqtt.MQTT_ERR_SUCCESS:
                result = True
                break
            else:
                time.sleep(self._PUBLISH_WAIT)
        return result

    def pubsub_receive(self, message):
        """
        Make a request/response using publishing and subscription

        Parameters
        ----------
        message: Request message

        Returns
        ----------
        result: Execution result
        rec_buff: Response message
        """
        result = False
        rec_buff = None
        # Initialization of received data storage variable
        self._lock.acquire()
        self._sub_receive_flag = False
        self._sub_receive_message = None
        self._lock.release()
        # Automatic failure if topic is unregistered
        if not self._sub_topic:
            return result, rec_buff
        # Publish request
        if self.publish_send(message):
            # Wait for response result for 5 seconds
            for wait_count in range(self._RETRY_PUBSUB_MAX):
                time.sleep(self._PUBSUB_WAIT)
                # Get message when response result is received
                if self._sub_receive_flag:
                    rec_buff = self._sub_receive_message
                    result = True
                    self._lock.acquire()
                    self._sub_receive_flag = False
                    self._sub_receive_message = None
                    self._lock.release()
                    break
        return result, rec_buff

    def on_connect(self, client, userdata, flag, rc):
        """
        Callback to be called when a connection with MQTT Server is established

        Parameters
        ----------
        client: the client instance for this callback
        userdata: the private user data
        flags: response flags sent by the broker
        rc: the connection result
        """
        logger.info("Connected with result code {}".format(rc))
        if rc == mqtt.MQTT_ERR_SUCCESS:
            self._sub_connect_flag = True
        else:
            self._sub_connect_flag = False

    def on_disconnect(self, client, userdata, flag, rc):
        """
        Callback called when MQTT disconnects

        Parameters
        ----------
        client: the client instance for this callback
        userdata: the private user data
        rc: the disconnection result
        """
        self.loop_stop()
        self._sub_connect_flag = False
        if rc != mqtt.MQTT_ERR_SUCCESS:
            logger.warning("Unexpected disconnection of {}".format(self._pub_topic))

    def on_message(self, client, userdata, msg):
        """
        Callback to execute when receiving a message from a subscribed topic

        Parameters
        ----------
        client: the client instance for this callback
        userdata: the private user data
        message: an instance of MQTTMessage.
        """
        # Convert message from Json to dictionary
        dict_message = json.loads(msg.payload)
        self._lock.acquire()
        self._sub_receive_flag = True
        self._sub_receive_message = dict_message
        self._lock.release()
        logger.debug("notification: {}".format(dict_message))
