#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-
import numpy as np
from scipy.spatial.transform import Rotation
import json
import configparser


class CameraCalibration:

    def __init__(self, file_path, identification_id):
        self.__int_matrix = {"fx": 0, "fy": 0, "ox": 0, "oy": 0}
        self.__int_dist = {"k1": 0, "k2": 0, "p1": 0, "p2": 0, "k3": 0}
        self.__ext_rot_vect = np.matrix([0, 0, 0])
        self.__ext_trns_vect = np.array([0, 0, 0])
        self.__camera_id = 0
        self.__path = file_path
        self.__identification_id = identification_id

    def calibration_analysis(self):
        """
        Analyze the calibration file to get
        the external and internal parameters.

        Returns
        --------------------
        Analysis result
        """
        ana_res = False
        try:
            with open(self.__path, 'r') as read_json:
                load_json = json.load(read_json)
                self.__int_matrix = {"fx": load_json["internal_params"]["fx"],
                                     "fy": load_json["internal_params"]["fy"],
                                     "ox": load_json["internal_params"]["cx"],
                                     "oy": load_json["internal_params"]["cy"]}
                self.__int_dist = {"k1": load_json["internal_params"]["k1"],
                                   "k2": load_json["internal_params"]["k2"],
                                   "p1": load_json["internal_params"]["p1"],
                                   "p2": load_json["internal_params"]["p2"],
                                   "k3": load_json["internal_params"]["k3"]}
                self.__ext_rot_vect = np.array([load_json["external_params"]["rx"],
                                                load_json["external_params"]["ry"],
                                                load_json["external_params"]["rz"]])
                self.__ext_trns_vect = np.array([[load_json["external_params"]["tx"]],
                                                 [load_json["external_params"]["ty"]],
                                                 [load_json["external_params"]["tz"]]])
                self.__camera_id = load_json["camera_id"]
                ana_res = True
        except json.JSONDecodeError:
            pass
        return ana_res

    def get_calibration_param(self):
        """
        Provides calibration information (external and internal parameters).

        Returns
        --------------------
        calibration information
        """
        return {
            "internal_matrix": self.__int_matrix,
            "internal_distortion": self.__int_dist,
            "external_rotation": self.__ext_rot_vect,
            "external_translation": self.__ext_trns_vect,
            "camera_id": self.__camera_id,
            "identification_id": self.__identification_id
        }


class ObjectPositionCalculator:

    def __init__(self, calibration: CameraCalibration):
        self.camera_calibration = calibration

    def coord_cnv_img_to_world(self, image_coord: list, class_name):
        """
        Converts image coordinates to world coordinates

        Parameters
        ----------
        image_coord: image coordinates
        class_name: Recognition target

        Returns
        --------------------
        World coordinates
        """
        if image_coord == []:
            return [None, None, None]

        u = image_coord[0]
        v = image_coord[1]

        # Store internal and external parameters
        camera_param = self.camera_calibration.get_calibration_param()
        fx = camera_param["internal_matrix"]["fx"]
        fy = camera_param["internal_matrix"]["fy"]
        cx = camera_param["internal_matrix"]["ox"]
        cy = camera_param["internal_matrix"]["oy"]
        rx = camera_param["external_rotation"][0]
        ry = camera_param["external_rotation"][1]
        rz = camera_param["external_rotation"][2]
        tx = camera_param["external_translation"][0]
        ty = camera_param["external_translation"][1]
        tz = camera_param["external_translation"][2]

        # Internal parameter matrix
        K = np.array([[fx, 0, cx], [0, fy, cy], [0, 0, 1]])

        # Rotation vector
        Rvec = np.array([rx, ry, rz])

        # Convert from rotation vector to rotation matrix
        rotational_inv_mat = Rotation.from_rotvec(Rvec)
        R = rotational_inv_mat.as_dcm()

        # Transition matrix
        T = np.array([tx, ty, tz])

        # screen pos
        cs = np.array([[u], [v], [1]])

        U = np.linalg.inv(R) @ T
        D = (np.linalg.inv(R) @ np.linalg.inv(K)) @ cs

        # Determine the target and get the Z coordinate value.
        Zconst = self.find_class_fixed_coord(class_name)
        if D[2][0] == 0:
            return [None, None, None]

        # Calculate the scaling factor
        s = (Zconst + U[2][0]) / (D[2][0])

        # Calculate world coordinetes
        calculate_coord = np.linalg.inv(R) @ (((np.linalg.inv(K) @ cs) * s) - T)
        world_coord = np.array([calculate_coord[0][0], calculate_coord[1][0], Zconst])
        return world_coord

    def find_class_fixed_coord(self, class_name):
        """
        Provides the 3D Z coordinate corresponding to the recognition target.

        Parameters
        ----------
        class_name: Recognition target

        Returns
        --------------------
        Z coordinate
        """
        Zconst = 0.5
        try:
            config_ini = configparser.ConfigParser()
            config_ini.read("./config/config.ini")
            if int(config_ini["coordinate"]["Z"]) == 0:
                return 0
            elif int(config_ini["coordinate"]["Z"]) == 1:
                return 0.0325
            with open("./config/object.json") as read_json:
                load_json = json.load(read_json)
                for obj in load_json:
                    if class_name == obj["class"]:
                        Zconst = obj["z-coordinate"]
                        break
        except json.JSONDecodeError:
            pass
        except OSError:
            pass
        return Zconst
