#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-
import base64
import datetime
import json
import logging
import os
from pytz import timezone
import requests
import time
from layout_monitoring import LayoutMonitoring
from mqtt_interface import MqttInterface
from object_position_calculator import ObjectPositionCalculator, CameraCalibration
from object_info import ImageInfo
from object_recognizer import ObjectRecognizer
from web_cam_interface import WebcamInterface

FORMAT = '%(asctime)s-%(levelname)s: %(message)s'
logging.basicConfig(level=logging.INFO, format=FORMAT)
logger = logging.getLogger(__name__)


class ObjectRecognitionController():

    def __init__(self):
        self._mqtt_host = "xxx.xxx.xxx.xxx" # set
        self._mqtt_port = 00000 # set
        self._upload_api_url = "http://RDBS/rdr/api/put/image_uploader" # set
        self._calibration_file_path = "./config/camera_conf.json"
        self._output_dir = "./tmp"
        self._calib_ins_a = []
        self._layout_count_time = 0
        self._mptt_pub_ins = None
        self._mqtt_topic = "obj_recognition_01/estimated_position"
        self._layout_target_cam = 2
        self.OBJECT_RECOGNITION_INTERVAL = 15
        self._LAYOUT_RUN_INTERVAL = 30
        self._HTTP_RESPONSE_OK = 200
        self._HTTP_RESPONSE_WAIT_SEC = 1
        self._RETRY_COUNT_MAX = 3
        self._layout_target_data = None
        self._camera_info_ins = None
        self._layout_ins = LayoutMonitoring()
        logger.info("ObjectRecognizer call")
        self._detecter = ObjectRecognizer()

    def run(self, signum, frame):
        """
        Main processing of object recognition system.
        Calculate the estimated position for each camera and
        write it to the DB. Layout change detection is also carried out.
        """
        logger.info("object recognition running...")
        image_path_a = []
        self._layout_target_data = None
        self._layout_count_time += self.OBJECT_RECOGNITION_INTERVAL
        # Object recognition for the number of cameras
        for calib_ins in self._calib_ins_a:
            # Get camera calibration information
            camera_param = calib_ins.get_calibration_param()
            image_info = ImageInfo()
            # Get camera capture
            camera_img = self._camera_info_ins.get_camera_capture(camera_param["camera_id"])
            if camera_img is None:
                break
            timestamp = datetime.datetime.now()
            ust_timestamp = timestamp.astimezone(timezone('UTC'))
            # Store camera information in structure
            image_info.shooting_time = (ust_timestamp.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]) + "+09:00"
            image_info.group_dir_path = self._output_dir
            image_info.name = 'camera_{}_{}.jpg'.format(camera_param["camera_id"], timestamp.strftime('%Y%m%d%H%M%S'))
            image_info.file_path = "{}{}{}".format(image_info.group_dir_path, '/', image_info.name)
            camera_img.save(image_info.file_path)
            image_path_a.append(image_info.file_path)

            # Object detection
            object_dict_list = self._detecter.object_detection(image_info.file_path)
            pos_cal_ins = ObjectPositionCalculator(calib_ins)

            for object_dict in object_dict_list:
                # Calculate the world coordinates of the detected object
                world_coord = pos_cal_ins.coord_cnv_img_to_world(
                    object_dict.image_coordinate, object_dict.recognition_name)
                object_dict.world_coordinate = world_coord
            image_info.object = object_dict_list
            # Upload camera capture
            send_file_path = self.image_upload(image_info.file_path)
            if send_file_path is None:
                logger.error("Unable to upload camera capture")
                for image_path in image_path_a:
                    os.remove(image_path)
                return
            # Upload estimated location information
            upload_res = self.object_position_upload(image_info, send_file_path, camera_param["camera_id"])
            if not upload_res:
                logger.error("Unable to update estimated position data for object")
                for image_path in image_path_a:
                    os.remove(image_path)
                return
            # Stores data used for layout change detection
            if camera_param["camera_id"] == self._layout_target_cam:
                self._layout_target_data = image_info.object
        # Layout execution interval judgment
        if self._LAYOUT_RUN_INTERVAL <= self._layout_count_time:
            if self._layout_target_data is not None:
                self._layout_count_time = 0
                # Layout change detection
                self._layout_ins.layout_change_detection(self._layout_target_data)
        # Delete the saved camera capture
        for image_path in image_path_a:
            os.remove(image_path)
        logger.info("object recognition end of cycle...")
        return

    def setting(self):
        """
        Perform the initial setting process.
        Analyze camera calibration information and
        connect with MQTT.

        Returns
        --------------------
        Initial setting execution result
        """
        result = False
        camera_id_a = []
        camera_ident_a = []
        # Camera calibration analysis
        if not self.camera_conf_analysis():
            return result
        for calib_ins in self._calib_ins_a:
            # Store camera ID and identification ID
            camera_param = calib_ins.get_calibration_param()
            camera_id_a.append(camera_param["camera_id"])
            camera_ident_a.append(camera_param["identification_id"])
        # Webcam initial settings
        self._camera_info_ins = WebcamInterface(camera_id_a, camera_ident_a)
        self._mptt_pub_ins = MqttInterface(self._mqtt_host, self._mqtt_port, self._mqtt_topic)
        # Connect to MQTT Server
        if not self._mptt_pub_ins.connection():
            return result
        if not self._layout_ins.mptt_req_res_ins.connection():
            return result
        if not self._layout_ins.mptt_pub_ins.connection():
            return result
        result = True
        return result

    def camera_conf_analysis(self):
        """
        Check and analyze input files

        Returns
        ----------------------------
        Check and analysis results
        """
        result = False
        try:
            with open(self._calibration_file_path, 'r') as conf_file:
                conf_json_list_a = json.load(conf_file)
                for conf_json in conf_json_list_a:
                    calibration_path = "{}/{}".format(
                        os.path.dirname(self._calibration_file_path), conf_json["Calibration"])
                    cal_ins = CameraCalibration(calibration_path, conf_json["CameraId"])
                    if cal_ins.calibration_analysis():
                        self._calib_ins_a.append(cal_ins)
                    else:
                        self._calib_ins_a = []
                        break
                else:
                    result = True
        except json.JSONDecodeError:
            pass
        except OSError:
            pass
        return result

    def image_upload(self, image_path):
        """
        Upload camera capture

        Parameters
        ----------------------------
        image_path: Local path of camera capture to upload

        Returns
        ----------------------------
        Path information of uploaded camera capture
        """
        result = None
        try:
            with open(image_path, 'rb') as img_data:
                enc_img = base64.b64encode(img_data.read()).decode('utf-8')
                send_file = {'robotId': "object_recognition_01", "imageBase64": enc_img}
                for retry_count in range(self._RETRY_COUNT_MAX):
                    post_res = requests.post(self._upload_api_url, json.dumps(send_file), headers={'Content-Type': 'application/json'}, timeout=(5.0))
                    if post_res.status_code == self._HTTP_RESPONSE_OK:
                        res_dict = json.loads(post_res.text)
                        result = res_dict["path"]
                        break
                    else:
                        time.sleep(self._HTTP_RESPONSE_WAIT_SEC)
        except json.JSONDecodeError:
            pass
        except OSError:
            pass
        return result

    def object_position_upload(self, image_data, upload_path, camera_id):
        """
        Upload estimated location information

        Parameters
        ----------------------------
        image_data: Structures related to images and objects
        upload_path: Path of uploaded image
        camera_id: Camera ID

        Returns
        ----------------------------
        Upload result
        """
        object_data_a = []
        # Convert data to Json format
        for object_data in image_data.object:
            object_dict = {
                "object_name": object_data.recognition_name,
                "detection_rate": float(object_data.detect_accuracy),
                "image_x": float(object_data.image_coordinate[0]),
                "image_y": float(object_data.image_coordinate[1]),
                "object_x": float(object_data.world_coordinate[0]),
                "object_y": float(object_data.world_coordinate[1]),
                "object_z": float(object_data.world_coordinate[2])
            }
            object_data_a.append(object_dict)

        send_json = {
            "robot_id": "object_recognition_01",
            "image_info": {
                "image_name": image_data.name,
                "image_date": image_data.shooting_time,
                "image_path": upload_path
            },
            "object_list": object_data_a,
            "camera_id": camera_id
        }
        return self._mptt_pub_ins.publish_send(send_json)


# if __name__ == "__main__":
#     ins = ObjectRecognitionController()
#     ins.setting()
#     ins.run(None, None)
