#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-
import cv2
import keras
import numpy as np
import logging

import tensorflow as tf

config = tf.ConfigProto()
config.gpu_options.allow_growth = True
tf.keras.backend.set_session(tf.Session(config=config))

from config import *
from model.decode_np import Decode
from model.yolo import *
from tools.cocotools import get_classes
from object_info import PerceptibleObject

FORMAT = '%(asctime)s-%(levelname)s: %(message)s'
logging.basicConfig(level=logging.INFO, format=FORMAT)
logger = logging.getLogger(__name__)


DEFAULT_SET_CONFIG = 1

YOLOV4_2X_CONFIG = 0
PPYOLO_2X_CONFIG = 1
PPYOLO_R18VD_CONFIG = 2


class ObjectRecognizer():
    def __init__(self):
        self.config_file = DEFAULT_SET_CONFIG
        self.cfg = None
        if self.config_file == YOLOV4_2X_CONFIG:
            self.cfg = YOLOv4_2x_Config()
        elif self.config_file == PPYOLO_2X_CONFIG:
            self.cfg = PPYOLO_2x_Config()
        elif self.config_file == PPYOLO_R18VD_CONFIG:
            self.cfg = PPYOLO_r18vd_Config()
        self.use_gpu = True
        self.cfg.test_cfg['model_path'] = "model/object_recognition_model.h5"
        self.cfg.classes_path = "model/object_recognition_class.txt"
        
        # Read model
        model_path = self.cfg.test_cfg['model_path']

        # Whether to frame the picture.
        self.draw_image = self.cfg.test_cfg['draw_image']
        self.draw_thresh = self.cfg.test_cfg['draw_thresh']

        self.all_classes = get_classes(self.cfg.classes_path)
        self.num_classes = len(self.all_classes)
        self.cfg.head["num_classes"] = self.num_classes
        self.cfg.gt2YoloTarget["num_classes"] = self.num_classes
        
        # Create model
        Backbone = select_backbone(self.cfg.backbone_type)
        backbone = Backbone(**self.cfg.backbone)
        Head = select_head(self.cfg.head_type)
        # Turn off DropBlock during prediction to obtain consistent inference results.
        self.cfg.head['drop_block'] = False
        head = Head(yolo_loss=None, nms_cfg=self.cfg.nms_cfg, **self.cfg.head)
        self.yolo = YOLO(backbone, head)

        x = keras.layers.Input(shape=(None, None, 3),
                               name='x', dtype='float32')
        self.im_size = keras.layers.Input(shape=(2,),
                                          name='im_size', dtype='int32')
        outputs = self.yolo.get_outputs(x)
        preds = self.yolo.get_prediction(outputs, self.im_size)
        predict_model = keras.models.Model(inputs=[x, self.im_size],
                                           outputs=preds)
        predict_model.load_weights(model_path,
                                   by_name=True,
                                   skip_mismatch=True)
        predict_model.summary(line_length=130)
        self._decode = Decode(predict_model,
                              self.all_classes,
                              self.use_gpu,
                              self.cfg,
                              for_test=False)
        logger.info("ObjectRecognizer call end")

    def object_detection(self, image_path):
        """
        Performs object recognition and returns detected object information

        Parameters
        ----------------------------
        image_path: image file path

        Returns
        --------------------
        Initial setting execution result
        """
        image = cv2.imread(image_path)
        pimage, im_size = self._decode.process_image(np.copy(image))
        dic = {}
        dic['image'] = image
        dic['pimage'] = pimage
        dic['im_size'] = im_size

        image, boxes, scores, classes = self._decode.detect_image(
            image, pimage, im_size, self.draw_image, self.draw_thresh)

        obj_list = []
        for box, score, cl in zip(boxes, scores, classes):
            if score >= self.draw_thresh:
                obj = PerceptibleObject()
                x0, y0, x1, y1 = box
                left = max(0, np.floor(x0 + 0.5).astype(int))
                # top = max(0, np.floor(y0 + 0.5).astype(int))
                right = min(image.shape[1], np.floor(x1 + 0.5).astype(int))
                bottom = min(image.shape[0], np.floor(y1 + 0.5).astype(int))
                class_name = self.all_classes[cl]
                obj.recognition_name = class_name
                obj.detect_accuracy = score
                pos_x = int((left + right) / 2)
                pos_y = bottom
                obj.image_coordinate = [pos_x, pos_y]
                obj_list.append(obj)
            else:
                continue

        #cv2.imwrite(image_path, image)
        return obj_list
